/*
 * Copyright (c) 1992, 1993, 1995
 *      The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software donated to Berkeley by
 * Jan-Simon Pendry.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)rot13fs_vfsops.c     8.2 (Berkeley) 1/21/94
 *
 * @(#)lofs_vfsops.c    1.2 (Berkeley) 6/18/92
 * $Id: vfs.c,v 1.1.1.1 1998/11/05 21:06:36 ezk Exp $
 */

/*
 * Rot13fs Layer
 * (See rot13fs_vnops.c for a description of what this does.)
 */

#include "opt_debug.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <rot13fs.h>

static MALLOC_DEFINE(M_ROT13FSMNT, "ROT13FS mount", "ROT13FS mount structure");

static int rot13fs_fhtovp __P((struct mount *mp, struct fid * fidp,
			      struct sockaddr * nam, struct vnode ** vpp,
			      int *exflagsp, struct ucred ** credanonp));
static int rot13fs_mount __P((struct mount *mp, char *path, caddr_t data,
			     struct nameidata * ndp, struct proc * p));
static int rot13fs_quotactl __P((struct mount *mp, int cmd, uid_t uid,
				caddr_t arg, struct proc * p));
static int rot13fs_root __P((struct mount *mp, struct vnode ** vpp));
static int rot13fs_start __P((struct mount *mp, int flags, struct proc * p));
static int rot13fs_statfs __P((struct mount *mp, struct statfs * sbp,
			      struct proc * p));
static int rot13fs_sync __P((struct mount *mp, int waitfor,
			    struct ucred * cred, struct proc * p));
static int rot13fs_unmount __P((struct mount *mp, int mntflags,
			       struct proc * p));
static int rot13fs_vget __P((struct mount *mp, ino_t ino,
			    struct vnode ** vpp));
static int rot13fs_vptofh __P((struct vnode * vp, struct fid * fhp));

/*
 * Mount rot13fs layer
 */
static int
rot13fs_mount(mp, path, data, ndp, p)
     struct mount *mp;
     char *path;
     caddr_t data;
     struct nameidata *ndp;
     struct proc *p;
{
  int error = 0;
  struct rot13fs_args args;
  struct vnode *lowerrootvp, *vp;
  struct vnode *rot13fsm_rootvp;
  struct rot13fs_mount *xmp;
  u_int size;
  int isvnunlocked = 0;

#ifdef ROT13FS_DIAGNOSTIC
  printf("rot13fs_mount(mp = %p)\n", mp);
#endif

  /*
   * Update is a no-op
   */
  if (mp->mnt_flag & MNT_UPDATE) {
    return (EOPNOTSUPP);
    /* return VFS_MOUNT(MOUNTTOROT13FSMOUNT(mp)->rot13fsm_vfs, path, data, ndp,
     * p); */
  }
  /*
   * Get argument
   */
  error = copyin(data, (caddr_t) & args, sizeof(struct rot13fs_args));
  if (error)
    return (error);

  /*
   * Unlock lower node to avoid deadlock.
   * (XXX) VOP_ISLOCKED is needed?
   */
  if ((mp->mnt_vnodecovered->v_op == rot13fs_vnodeop_p) &&
      VOP_ISLOCKED(mp->mnt_vnodecovered)) {
    VOP_UNLOCK(mp->mnt_vnodecovered, 0, p);
    isvnunlocked = 1;
  }
  /*
   * Find lower node
   */
  NDINIT(ndp, LOOKUP, FOLLOW | WANTPARENT | LOCKLEAF,
	 UIO_USERSPACE, args.target, p);
  error = namei(ndp);
  /*
   * Re-lock vnode.
   */
  if (isvnunlocked && !VOP_ISLOCKED(mp->mnt_vnodecovered))
    vn_lock(mp->mnt_vnodecovered, LK_EXCLUSIVE | LK_RETRY, p);

  if (error)
    return (error);

  /*
   * Sanity check on lower vnode
   */
  lowerrootvp = ndp->ni_vp;

  vrele(ndp->ni_dvp);
  ndp->ni_dvp = NULLVP;

  /*
   * Check multi rot13fs mount to avoid `lock against myself' panic.
   */
  if (lowerrootvp == VP_TO_ROT13FS(mp->mnt_vnodecovered)->rot13fs_lowervp) {
#ifdef DIAGNOSTIC
    printf("rot13fs_mount: multi rot13fs mount?\n");
#endif
    return (EDEADLK);
  }
  xmp = (struct rot13fs_mount *) malloc(sizeof(struct rot13fs_mount),
				       M_ROT13FSMNT, M_WAITOK);	/* XXX */

  /*
   * Save reference to underlying FS
   */
  xmp->rot13fsm_vfs = lowerrootvp->v_mount;

  /*
   * Save reference.  Each mount also holds
   * a reference on the root vnode.
   */
  error = rot13fs_node_create(mp, lowerrootvp, &vp);
  /*
   * Unlock the node (either the lower or the alias)
   */
  VOP_UNLOCK(vp, 0, p);
  /*
   * Make sure the node alias worked
   */
  if (error) {
    vrele(lowerrootvp);
    free(xmp, M_ROT13FSMNT);	/* XXX */
    return (error);
  }
  /*
   * Keep a held reference to the root vnode.
   * It is vrele'd in rot13fs_unmount.
   */
  rot13fsm_rootvp = vp;
  rot13fsm_rootvp->v_flag |= VROOT;
  xmp->rot13fsm_rootvp = rot13fsm_rootvp;
  if (ROT13FS_VP_TO_LOWERVP(rot13fsm_rootvp)->v_mount->mnt_flag & MNT_LOCAL)
    mp->mnt_flag |= MNT_LOCAL;
  mp->mnt_data = (qaddr_t) xmp;
  vfs_getnewfsid(mp);

  (void) copyinstr(path, mp->mnt_stat.f_mntonname, MNAMELEN - 1, &size);
  bzero(mp->mnt_stat.f_mntonname + size, MNAMELEN - size);
  (void) copyinstr(args.target, mp->mnt_stat.f_mntfromname, MNAMELEN - 1,
		   &size);
  bzero(mp->mnt_stat.f_mntfromname + size, MNAMELEN - size);
  (void) rot13fs_statfs(mp, &mp->mnt_stat, p);
#ifdef ROT13FS_DIAGNOSTIC
  printf("rot13fs_mount: lower %s, alias at %s\n",
	 mp->mnt_stat.f_mntfromname, mp->mnt_stat.f_mntonname);
#endif
  return (0);
}

/*
 * VFS start.  Nothing needed here - the start routine
 * on the underlying filesystem will have been called
 * when that filesystem was mounted.
 */
static int
rot13fs_start(mp, flags, p)
     struct mount *mp;
     int flags;
     struct proc *p;
{
  return (0);
  /* return VFS_START(MOUNTTOROT13FSMOUNT(mp)->rot13fsm_vfs, flags, p); */
}

/*
 * Free reference to rot13fs layer
 */
static int
rot13fs_unmount(mp, mntflags, p)
     struct mount *mp;
     int mntflags;
     struct proc *p;
{
  struct vnode *rot13fsm_rootvp = MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_rootvp;
  int error;
  int flags = 0;

#ifdef ROT13FS_DIAGNOSTIC
  printf("rot13fs_unmount(mp = %p)\n", mp);
#endif

  if (mntflags & MNT_FORCE)
    flags |= FORCECLOSE;

  /*
   * Clear out buffer cache.  I don't think we
   * ever get anything cached at this level at the
   * moment, but who knows...
   */
#if 0
  mntflushbuf(mp, 0);
  if (mntinvalbuf(mp, 1))
    return (EBUSY);
#endif
  if (rot13fsm_rootvp->v_usecount > 1)
    return (EBUSY);
  error = vflush(mp, rot13fsm_rootvp, flags);
  if (error)
    return (error);

#ifdef ROT13FS_DIAGNOSTIC
  vprint("alias root of lower", rot13fsm_rootvp);
#endif
  /*
   * Release reference on underlying root vnode
   */
  vrele(rot13fsm_rootvp);
  /*
   * And blow it away for future re-use
   */
  vgone(rot13fsm_rootvp);
  /*
   * Finally, throw away the rot13fs_mount structure
   */
  free(mp->mnt_data, M_ROT13FSMNT);	/* XXX */
  mp->mnt_data = 0;
  return 0;
}

static int
rot13fs_root(mp, vpp)
     struct mount *mp;
     struct vnode **vpp;
{
  struct proc *p = curproc;	/* XXX */
  struct vnode *vp;

#ifdef ROT13FS_DIAGNOSTIC
  printf("rot13fs_root(mp = %p, vp = %x->%x)\n", mp,
	 MOUNTTOROT13FSMOUNT(mp)->rot13fsm_rootvp,
	 ROT13FSVPTOLOWERVP(MOUNTTOROT13FSMOUNT(mp)->rot13fsm_rootvp)
    );
#endif

  /*
   * Return locked reference to root.
   */
  vp = MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_rootvp;
  VREF(vp);
  if (VOP_ISLOCKED(vp)) {
    /*
     * XXX
     * Should we check type of node?
     */
#ifdef DIAGNOSTIC
    printf("rot13fs_root: multi rot13fs mount?\n");
#endif
    vrele(vp);
    return (EDEADLK);
  } else
    vn_lock(vp, LK_EXCLUSIVE | LK_RETRY, p);
  *vpp = vp;
  return 0;
}

static int
rot13fs_quotactl(mp, cmd, uid, arg, p)
     struct mount *mp;
     int cmd;
     uid_t uid;
     caddr_t arg;
     struct proc *p;
{
  return VFS_QUOTACTL(MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_vfs, cmd, uid, arg, p);
}

static int
rot13fs_statfs(mp, sbp, p)
     struct mount *mp;
     struct statfs *sbp;
     struct proc *p;
{
  int error;
  struct statfs mstat;

#ifdef ROT13FS_DIAGNOSTIC
  printf("rot13fs_statfs(mp = %p, vp = %x->%x)\n", mp,
	 MOUNTTOROT13FSMOUNT(mp)->rot13fsm_rootvp,
	 ROT13FSVPTOLOWERVP(MOUNTTOROT13FSMOUNT(mp)->rot13fsm_rootvp)
    );
#endif

  bzero(&mstat, sizeof(mstat));

  error = VFS_STATFS(MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_vfs, &mstat, p);
  if (error)
    return (error);

  /* now copy across the "interesting" information and fake the rest */
  sbp->f_type = mstat.f_type;
  sbp->f_flags = mstat.f_flags;
  sbp->f_bsize = mstat.f_bsize;
  sbp->f_iosize = mstat.f_iosize;
  sbp->f_blocks = mstat.f_blocks;
  sbp->f_bfree = mstat.f_bfree;
  sbp->f_bavail = mstat.f_bavail;
  sbp->f_files = mstat.f_files;
  sbp->f_ffree = mstat.f_ffree;
  if (sbp != &mp->mnt_stat) {
    bcopy(&mp->mnt_stat.f_fsid, &sbp->f_fsid, sizeof(sbp->f_fsid));
    bcopy(mp->mnt_stat.f_mntonname, sbp->f_mntonname, MNAMELEN);
    bcopy(mp->mnt_stat.f_mntfromname, sbp->f_mntfromname, MNAMELEN);
  }
  return (0);
}

static int
rot13fs_sync(mp, waitfor, cred, p)
     struct mount *mp;
     int waitfor;
     struct ucred *cred;
     struct proc *p;
{
  /*
   * XXX - Assumes no data cached at rot13fs layer.
   */
  fist_dprint(4, "ROT13FS_SYNC: vfs=0x%x\n", (int) mp);
  return (0);
}

static int
rot13fs_vget(mp, ino, vpp)
     struct mount *mp;
     ino_t ino;
     struct vnode **vpp;
{

  return VFS_VGET(MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_vfs, ino, vpp);
}

static int
rot13fs_fhtovp(mp, fidp, nam, vpp, exflagsp, credanonp)
     struct mount *mp;
     struct fid *fidp;
     struct sockaddr *nam;
     struct vnode **vpp;
     int *exflagsp;
     struct ucred **credanonp;
{

  return VFS_FHTOVP(MOUNT_TO_ROT13FS_MOUNT(mp)->rot13fsm_vfs, fidp, nam,
		    vpp, exflagsp, credanonp);
}

static int
rot13fs_vptofh(vp, fhp)
     struct vnode *vp;
     struct fid *fhp;
{
  return VFS_VPTOFH(ROT13FS_VP_TO_LOWERVP(vp), fhp);
}

static struct vfsops rot13fs_vfsops =
{
  rot13fs_mount,
  rot13fs_start,
  rot13fs_unmount,
  rot13fs_root,
  rot13fs_quotactl,
  rot13fs_statfs,
  rot13fs_sync,
  rot13fs_vget,
  rot13fs_fhtovp,
  rot13fs_vptofh,
  rot13fs_init,
};

VFS_SET(rot13fs_vfsops, rot13fs, VFCF_LOOPBACK);
